# THIS MODULE IS FOR BACKWARDS COMPATIBILITY ONLY!
from collections.abc import Iterable

from sybil import Region, Document
from sybil.parsers.rest import CaptureParser


def parse_captures(document: Document) -> Iterable[Region]:
    """
    A parser function to be included when your documentation makes use of
    :ref:`capture-parser` examples.
    """
    return CaptureParser()(document)



from typing import TYPE_CHECKING, Any, Dict

from .region import Region

if TYPE_CHECKING:
    from .document import Document


class SybilFailure(AssertionError):

    def __init__(self, example: 'Example', result: str) -> None:
        super(SybilFailure, self).__init__((
            'Example at {}, line {}, column {} did not evaluate as expected:\n'
            '{}'
        ).format(example.path, example.line, example.column, result))
        self.example = example
        self.result = result


class NotEvaluated(Exception):
    """
    An exception that can be raised by an :any:`Evaluator` previously
    :meth:`pushed <sybil.Document.push_evaluator>` onto the document to indicate that
    it is not evaluating the current example and that a previously pushed evaluator, or the
    :any:`Region` evaluator if no others have been pushed, should be used to evaluate the
    :any:`Example` instead.
    """


class Example:
    """
    This represents a particular example from a documentation source file.
    It is assembled from the :class:`~sybil.document.Document` and 
    :class:`~sybil.Region` the example comes from and is passed to the region's 
    evaluator.
    """

    def __init__(
        self, document: 'Document', line: int, column: int, region: Region, namespace: Dict[str, Any]
    ) -> None:
        #: The :class:`~sybil.document.Document` from which this example came.
        self.document: 'Document' = document
        #: The absolute path of the :class:`~sybil.document.Document`.
        self.path: str = document.path
        #: The line number at which this example occurs in the
        #: :class:`~sybil.document.Document`.
        self.line: int = line
        #: The column number at which this example occurs in the
        #: :class:`~sybil.document.Document`.
        self.column: int = column
        #: The :class:`~sybil.Region` from which this example came.
        self.region: Region = region
        #: The character position at which this example starts in the
        #: :class:`~sybil.document.Document`.
        self.start: int = region.start
        #: The character position at which this example ends in the
        #: :class:`~sybil.document.Document`.
        self.end: int = region.end
        #: The version of this example provided by the parser that yielded
        #: the :class:`~sybil.Region` containing it.
        self.parsed: Any = region.parsed
        #: The :attr:`~sybil.Document.namespace` of the document from
        #: which this example came.
        self.namespace: Dict[str, Any] = namespace

    def __repr__(self) -> str:
        return '<Example path={} line={} column={} using {!r}>'.format(
            self.path, self.line, self.column, self.region.evaluator
        )

    def evaluate(self) -> None:
        if self.region.evaluator is not None:
            self.document.evaluate(self, self.region.evaluator)



from .sybil import Sybil
from .document import Document
from .region import Region, Lexeme
from .example import Example

__all__ = [
    'Sybil',
    'Document',
    'Region',
    'Lexeme',
    'Example',
]



import re
from collections.abc import Iterable
from typing import Optional, Dict, Pattern, Match, List

from sybil import Document, Region, Lexeme
from sybil.parsers.abstract.lexers import BlockLexer, strip_prefix

FENCE = re.compile(r"^(?P<prefix>[ \t]*)(?P<fence>`{3,}|~{3,})", re.MULTILINE)


class RawFencedCodeBlockLexer:
    """
    A :class:`~sybil.typing.Lexer` for Markdown fenced code blocks allowing flexible lexing
    of the whole `info` line along with more complicated prefixes.

    The following lexemes are extracted:

    - ``source`` as a :class:`~sybil.Lexeme`.
    - any other named groups specified in ``info_pattern`` as :class:`strings <str>`.


    :param info_pattern:
        a :class:`re.Pattern` to match the `info` line and any required prefix that follows it.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their
        values. Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.

    """


    def __init__(
            self,
            info_pattern: Pattern[str] = re.compile(r'$\n', re.MULTILINE),
            mapping: Optional[Dict[str, str]] = None,
    ) -> None:
        self.info_pattern = info_pattern
        self.mapping = mapping

    @staticmethod
    def match_closes_existing(current: Match[str], existing: Match[str]) -> bool:
        current_fence = current.group('fence')
        existing_fence = existing.group('fence')
        same_type = current_fence[0] == existing_fence[0]
        okay_length = len(current_fence) >= len(existing_fence)
        same_prefix = len(current.group('prefix')) == len(existing.group('prefix'))
        return same_type and okay_length and same_prefix

    def make_region(
            self, opening: Match[str], document: Document, closing: Optional[Match[str]]
    ) -> Optional[Region]:
        if closing is None:
            content_end = region_end = len(document.text)
        else:
            content_end = closing.start()
            region_end = closing.end()
        content = document.text[opening.end(): content_end]
        info = self.info_pattern.match(content)
        if info is None:
            return None
        lexemes = info.groupdict()
        lexemes['source'] = Lexeme(
            strip_prefix(content[info.end():], opening.group('prefix')),
            offset=len(opening.group(0))+info.end(),
            line_offset=0,
        )
        if self.mapping:
            lexemes = {dest: lexemes[source] for source, dest in self.mapping.items()}
        return Region(opening.start(), region_end, lexemes=lexemes)

    def __call__(self, document: Document) -> Iterable[Region]:
        open_blocks: List[Match[str]] = []
        index = 0
        while True:
            match = FENCE.search(document.text, index)
            if match is None:
                break
            else:
                index = match.end()
            # does this fence close any open block?
            for i in range(len(open_blocks)):
                existing = open_blocks[i]
                if self.match_closes_existing(match, existing):
                    maybe_region = self.make_region(existing, document, match)
                    if maybe_region is not None:
                        yield maybe_region
                    open_blocks = open_blocks[:i]
                    break
            else:
                open_blocks.append(match)
        if open_blocks:
            maybe_region = self.make_region(open_blocks[0], document, closing=None)
            if maybe_region is not None:
                yield maybe_region


class FencedCodeBlockLexer(RawFencedCodeBlockLexer):
    """
    A :class:`~sybil.typing.Lexer` for Markdown fenced code blocks where a language is specified.
    :class:`RawFencedCodeBlockLexer` can be used if the whole `info` line, or a more complicated
    prefix, is required.

    The following lexemes are extracted:

    - ``language`` as a  :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.


    :param language:
        a :class:`str` containing a regular expression pattern to match language names.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their
        values. Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.

    """

    def __init__(self, language: str, mapping: Optional[Dict[str, str]] = None) -> None:
        super().__init__(
            info_pattern=re.compile(f'(?P<language>{language})$\n', re.MULTILINE),
            mapping=mapping,
        )


DIRECTIVE_IN_HTML_COMMENT_START = (
    r"^(?P<prefix>[ \t]*)<!--+\s*(?:;\s*)?(?P<directive>{directive}):?[ \t]"
    r"*(?P<arguments>{arguments})[ \t]*"
    r"(?:$\n|(?=--+>))"
)
DIRECTIVE_IN_HTML_COMMENT_END = '(?:(?<=\n){prefix})?--+>'


class DirectiveInHTMLCommentLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for faux directives in
    HTML-style Markdown comments such as:

    .. code-block:: markdown

        <!--- not-really-a-directive: some-argument

            Source here...

        --->

    It extracts the following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.
    """

    def __init__(
            self, directive: str, arguments: str = '.*?', mapping: Optional[Dict[str, str]] = None
    ) -> None:
        super().__init__(
            start_pattern=re.compile(
                DIRECTIVE_IN_HTML_COMMENT_START.format(directive=directive, arguments=arguments),
                re.MULTILINE
            ),
            end_pattern_template=DIRECTIVE_IN_HTML_COMMENT_END,
            mapping=mapping,
        )



# THIS MODULE IS FOR BACKWARDS COMPATIBILITY ONLY!
from collections.abc import Iterable

from sybil import Region, Document
from .rest import SkipParser


def skip(document: Document) -> Iterable[Region]:
    """
    A parser function to be included when your documentation makes use of
    :ref:`skipping <skip-parser>` examples in a document.
    """
    return SkipParser()(document)



from dataclasses import dataclass
from typing import Any, Optional, Dict
from unittest import SkipTest

from sybil import Example, Document
from sybil.example import NotEvaluated


class If:

    def __init__(self, default_reason: str) -> None:
        self.default_reason = default_reason

    def __call__(self, condition: Any, reason: Optional[str] = None) -> Optional[str]:
        if condition:
            return reason or self.default_reason
        return None


@dataclass
class SkipState:
    active: bool = True
    remove: bool = False
    exception: Optional[Exception] = None
    last_action: Optional[str] = None


class Skipper:

    def __init__(self) -> None:
        self.document_state: Dict[Document, SkipState] = {}

    def state_for(self, example: Example) -> SkipState:
        document = example.document
        if document not in self.document_state:
            self.document_state[document] = SkipState()
        return self.document_state[example.document]

    def install(self, example: Example, state: SkipState, reason: Optional[str]) -> None:
        document = example.document
        document.push_evaluator(self)
        if reason:
            namespace = document.namespace.copy()
            reason = reason.lstrip()
            if reason.startswith('if'):
                condition = reason[2:]
                reason = 'if_' + condition
                namespace['if_'] = If(condition)
            reason = eval(reason, namespace)
            if reason:
                state.exception = SkipTest(reason)
            else:
                state.active = False

    def remove(self, example: Example) -> None:
        document = example.document
        document.pop_evaluator(self)
        del self.document_state[document]

    def evaluate_skip_example(self, example: Example) -> None:
        state = self.state_for(example)
        action, reason = example.parsed

        if action not in ('start', 'next', 'end'):
            raise ValueError('Bad skip action: ' + action)
        if state.last_action is None and action not in ('start', 'next'):
            raise ValueError(f"'skip: {action}' must follow 'skip: start'")
        elif state.last_action and action != 'end':
            raise ValueError(f"'skip: {action}' cannot follow 'skip: {state.last_action}'")

        state.last_action = action

        if action == 'start':
            self.install(example, state, reason)
        elif action == 'next':
            self.install(example, state, reason)
            state.remove = True
        elif action == 'end':
            self.remove(example)
            if reason:
                raise ValueError("Cannot have condition on 'skip: end'")

    def evaluate_other_example(self, example: Example) -> None:
        state = self.state_for(example)
        if state.remove:
            self.remove(example)
        if not state.active:
            raise NotEvaluated()
        if state.exception is not None:
            raise state.exception

    def __call__(self, example: Example) -> None:
        if example.region.evaluator is self:
            self.evaluate_skip_example(example)
        else:
            self.evaluate_other_example(example)



from __future__ import absolute_import

import os
from collections.abc import Callable, Sequence
from inspect import getsourcefile
from os.path import abspath
from pathlib import Path
from typing import Union, Tuple, Optional, List

import pytest
from pytest import Collector, ExceptionInfo, Module, Session
from _pytest import fixtures
from _pytest._code.code import TerminalRepr, Traceback
from _pytest._io import TerminalWriter
from _pytest.fixtures import FuncFixtureInfo

from sybil import example as example_module, Sybil, Document
from sybil.example import Example
from sybil.example import SybilFailure

PYTEST_VERSION = tuple(int(i) for i in pytest.__version__.split('.'))

example_module_path = abspath(getsourcefile(example_module))


class SybilFailureRepr(TerminalRepr):

    def __init__(self, item: 'SybilItem', message: str) -> None:
        self.item = item
        self.message = message

    def toterminal(self, tw: TerminalWriter) -> None:
        tw.line()
        for line in self.message.splitlines():
            tw.line(line)
        tw.line()
        tw.write(self.item.parent.name, bold=True, red=True)
        tw.line(":%s: SybilFailure" % self.item.example.line)


class SybilItem(pytest.Item):

    obj = None

    def __init__(self, parent, sybil, example: Example) -> None:
        super(SybilItem, self).__init__(sybil.identify(example), parent)
        self.example = example
        self.request_fixtures(sybil.fixtures)

    def request_fixtures(self, names):
        # pytest fixtures dance:
        fm = self.session._fixturemanager
        closure = fm.getfixtureclosure(initialnames=names, parentnode=self, ignore_args=set())
        names_closure, arg2fixturedefs = closure
        fixtureinfo = FuncFixtureInfo(argnames=names, initialnames=names, names_closure=names_closure, name2fixturedefs=arg2fixturedefs)
        self._fixtureinfo = fixtureinfo
        self.funcargs = {}
        self._request = fixtures.TopRequest(pyfuncitem=self, _ispytest=True)
        self.fixturenames = names_closure

    def reportinfo(self) -> Tuple[Union["os.PathLike[str]", str], Optional[int], str]:
        info = '%s line=%i column=%i' % (
            self.path.name, self.example.line, self.example.column
        )
        return self.example.path, self.example.line, info

    def getparent(self, cls):
        if cls is Module:
            return self.parent
        if cls is Session:
            return self.session

    def setup(self) -> None:
        self._request._fillfixtures()
        for name, fixture in self.funcargs.items():
            self.example.namespace[name] = fixture

    def runtest(self) -> None:
        self.example.evaluate()

    def _traceback_filter(self, excinfo: ExceptionInfo[BaseException]) -> Traceback:
        traceback = excinfo.traceback
        tb = traceback.cut(path=example_module_path)
        tb_entry = tb[1]
        if getattr(tb_entry, '_rawentry', None) is not None:
            traceback = Traceback(tb_entry._rawentry)
        return traceback

    def repr_failure(
        self,
        excinfo: ExceptionInfo[BaseException],
        style = None,
    ) -> Union[str, TerminalRepr]:
        if isinstance(excinfo.value, SybilFailure):
            return SybilFailureRepr(self, str(excinfo.value))
        return super().repr_failure(excinfo, style)


class SybilFile(pytest.File):

    def __init__(self, *, sybils: Sequence[Sybil], **kwargs) -> None:
        super(SybilFile, self).__init__(**kwargs)
        self.sybils: Sequence[Sybil] = sybils
        self.documents: List[Document] = []

    def collect(self):
        for sybil in self.sybils:
            document = sybil.parse(self.path)
            self.documents.append(document)
            for example in document.examples():
                yield SybilItem.from_parent(
                    self,
                    sybil=sybil,
                    example=example,
                )

    def setup(self) -> None:
        for sybil, document in zip(self.sybils, self.documents):
            if sybil.setup:
                sybil.setup(document.namespace)

    def teardown(self) -> None:
        for sybil, document in zip(self.sybils, self.documents):
            if sybil.teardown:
                sybil.teardown(document.namespace)


def pytest_integration(*sybils: Sybil) -> Callable[[Path, Collector], Optional[SybilFile]]:

    def pytest_collect_file(file_path: Path, parent: Collector) -> Optional[SybilFile]:
        active_sybils = [sybil for sybil in sybils if sybil.should_parse(file_path)]
        if active_sybils:
            return SybilFile.from_parent(parent, path=file_path, sybils=active_sybils)
        return None

    return pytest_collect_file



from typing import Optional

from sybil.parsers.abstract import AbstractCodeBlockParser
from sybil.typing import Evaluator
from .lexers import (
    DirectiveLexer, DirectiveInPercentCommentLexer
)
from ..markdown.lexers import FencedCodeBlockLexer, DirectiveInHTMLCommentLexer
from ..abstract.codeblock import PythonDocTestOrCodeBlockParser


class CodeBlockParser(AbstractCodeBlockParser):
    """
    A :any:`Parser` for :ref:`myst-codeblock-parser` examples.

    :param language:
        The language that this parser should look for.

    :param evaluator:
        The evaluator to use for evaluating code blocks in the specified language.
        You can also override the :meth:`evaluate` method below.
    """

    def __init__(
            self, language: Optional[str] = None, evaluator: Optional[Evaluator] = None
    ) -> None:
        super().__init__(
            [
                FencedCodeBlockLexer(
                    language=r'.+',
                    mapping={'language': 'arguments', 'source': 'source'},
                ),
                DirectiveLexer(
                    directive=r'(sourcecode|code-block|code)',
                    arguments='.+',
                ),
                DirectiveInPercentCommentLexer(
                    directive=r'(invisible-)?code(-block)?',
                    arguments='.+',
                ),
                DirectiveInHTMLCommentLexer(
                    directive=r'(invisible-)?code(-block)?',
                    arguments='.+',
                ),
            ],
            language, evaluator
        )


class PythonCodeBlockParser(PythonDocTestOrCodeBlockParser):
    """
    A :any:`Parser` for Python :ref:`myst-codeblock-parser` examples.

    :param future_imports:
        An optional list of strings that will be turned into
        ``from __future__ import ...`` statements and prepended to the code
        in each of the examples found by this parser.

    :param doctest_optionflags:
        :ref:`doctest option flags<option-flags-and-directives>` to use
        when evaluating the doctest examples found by this parser.
    """

    codeblock_parser_class = CodeBlockParser



from collections.abc import Callable
from typing import Any, Dict, Optional
from unittest import TestCase as BaseTestCase, TestSuite
from unittest.loader import TestLoader

from sybil import Sybil
from sybil.example import Example


class TestCase(BaseTestCase):

    sybil: Sybil
    namespace: Dict[str, Any]

    def __init__(self, example: Example) -> None:
        BaseTestCase.__init__(self)
        self.example = example

    def runTest(self) -> None:
        self.example.evaluate()

    def id(self) -> str:
        return f'{self.example.path},{self.sybil.identify(self.example)}'

    __str__ = __repr__ = id

    @classmethod
    def setUpClass(cls) -> None:
        if cls.sybil.setup is not None:
            cls.sybil.setup(cls.namespace)

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.sybil.teardown is not None:
            cls.sybil.teardown(cls.namespace)


def unittest_integration(
    *sybils: Sybil,
) -> Callable[[Optional[TestLoader], Optional[TestSuite], Optional[str]], TestSuite]:

    def load_tests(
        loader: Optional[TestLoader] = None,
        tests: Optional[TestSuite] = None,
        pattern: Optional[str] = None,
    ) -> TestSuite:
        suite = TestSuite()
        for sybil in sybils:
            for path in sorted(sybil.path.glob('**/*')):
                if path.is_file() and sybil.should_parse(path):
                    document = sybil.parse(path)

                    case = type(document.path, (TestCase, ), dict(
                        sybil=sybil, namespace=document.namespace,
                    ))

                    for example in document.examples():
                        suite.addTest(case(example))

        return suite

    return load_tests



from doctest import (
    DocTest as BaseDocTest,
    DocTestRunner as BaseDocTestRunner,
    Example as BaseDocTestExample,
    set_unittest_reportflags,
)
from typing import Any, Dict, List, Optional

from sybil import Example


class DocTest(BaseDocTest):
    def __init__(
            self,
            examples: List[BaseDocTestExample],
            globs: Dict[str, Any],
            name: str,
            filename: Optional[str],
            lineno: Optional[int],
            docstring: Optional[str],
        ) -> None:
        # do everything like regular doctests, but don't make a copy of globs
        BaseDocTest.__init__(self, examples, globs, name, filename, lineno, docstring)
        self.globs = globs


class DocTestRunner(BaseDocTestRunner):

    def __init__(self, optionflags: int) -> None:
        _unittest_reportflags = set_unittest_reportflags(0)
        set_unittest_reportflags(_unittest_reportflags)
        optionflags |= _unittest_reportflags
        BaseDocTestRunner.__init__(

            self,
            verbose=False,
            optionflags=optionflags,
        )

    def _failure_header(self, test: DocTest, example: BaseDocTestExample) -> str:
        return ''


class DocTestEvaluator:
    """
    The :any:`Evaluator` to use for :class:`Regions <sybil.Region>` yielded by
    a :class:`~sybil.parsers.abstract.doctest.DocTestStringParser`.


    :param optionflags:
        :ref:`doctest option flags<option-flags-and-directives>` to use
        when evaluating examples.
    """

    def __init__(self, optionflags: int = 0) -> None:
        self.runner = DocTestRunner(optionflags)

    def __call__(self, sybil_example: Example) -> str:
        example = sybil_example.parsed
        namespace = sybil_example.namespace
        output: List[str] = []
        remove_name = False
        try:
            if '__name__' not in namespace:
                remove_name = True
                namespace['__name__'] = '__test__'
            self.runner.run(
                DocTest([example], namespace, name=sybil_example.path,
                        filename=None, lineno=example.lineno, docstring=None),
                clear_globs=False,
                out=output.append
            )
        finally:
            if remove_name:
                del namespace['__name__']
        return ''.join(output)



from ..abstract import AbstractSkipParser
from .lexers import DirectiveInCommentLexer


class SkipParser(AbstractSkipParser):
    """
    A :any:`Parser` for :ref:`skip <skip-parser>` instructions.
    """

    def __init__(self) -> None:
        super().__init__([DirectiveInCommentLexer('skip')])



from collections.abc import Iterable

from sybil import Document, Region
from sybil.evaluators.doctest import DocTestEvaluator
from sybil.parsers.abstract import DocTestStringParser
from .lexers import DirectiveLexer


class DocTestDirectiveParser:
    """
    A :any:`Parser` for :ref:`doctest directive <myst-doctest-parser>` examples.

    :param optionflags:
        :ref:`doctest option flags<option-flags-and-directives>` to use
        when evaluating the examples found by this parser.

    """
    def __init__(self, optionflags: int = 0) -> None:
        self.lexer = DirectiveLexer('doctest')
        self.string_parser = DocTestStringParser(DocTestEvaluator(optionflags))

    def __call__(self, document: Document) -> Iterable[Region]:
        for lexed_region in self.lexer(document):
            source = lexed_region.lexemes['source']
            for region in self.string_parser(source, document.path):
                region.adjust(lexed_region, source)
                yield region



import re
from collections.abc import Iterable
from typing import Optional, Dict

from sybil import Document, Region
from sybil.parsers.abstract.lexers import BlockLexer

START_PATTERN_TEMPLATE =(
    r'^(?P<prefix>[ \t]*)\.\.\s*(?P<directive>{directive})'
    r'{delimiter}[ \t]*'
    r'(?P<arguments>[^\n]+)?\n'
    r'(?P<options>(?:\1[ \t]+:[\w-]*:[^\n]*\n)+)?'
)

OPTIONS_PATTERN = re.compile(r'[^:]*:(?P<name>[^:]+):[ \t]*(?P<value>[^\n]*)\n')
END_PATTERN_TEMPLATE = r'((?<=\n)(?=\.\.)|\n?\Z|\n[ \t]{{0,{len_prefix}}}(?=\S|\Z))'


def parse_options_and_source(lexed: Region) -> None:
    lexemes = lexed.lexemes
    raw_options = lexemes.pop('options', None)
    options = lexemes['options'] = {}
    if raw_options:
        for match in OPTIONS_PATTERN.finditer(raw_options):
            options[match['name']] = match['value']
    source = lexemes.get('source')
    if source:
        lexemes['source'] = source.strip_leading_newlines()


class DirectiveLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for ReST directives that extracts the
    following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.
    """

    delimiter = '::'

    def __init__(
            self,
            directive: str,
            arguments: str = '',
            mapping: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        A lexer for ReST directives.
        Both ``directive`` and ``arguments`` are regex patterns.
        """
        super().__init__(
            start_pattern=re.compile(
                START_PATTERN_TEMPLATE.format(
                    directive=directive,
                    delimiter=self.delimiter,
                    arguments=arguments
                ),
                re.MULTILINE
            ),
            end_pattern_template=END_PATTERN_TEMPLATE,
            mapping=mapping,
        )

    def __call__(self, document: Document) -> Iterable[Region]:
        for lexed in super().__call__(document):
            parse_options_and_source(lexed)
            yield lexed


class DirectiveInCommentLexer(DirectiveLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for faux ReST directives in comments
    such as:

    .. code-block:: rest

        .. not-really-a-directive: some-argument

          Source here...

    It extracts the following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.
    """

    # This is the pattern used for invisible code blocks and the like.
    delimiter = ':?'



import inspect
from pathlib import Path
from collections.abc import Callable, Collection, Mapping, Sequence
from typing import Any, Dict, Optional, Type, List, Tuple

from .document import Document, PythonDocStringDocument
from .example import Example
from .typing import Parser

DEFAULT_DOCUMENT_TYPES = {
    None: Document,
    '.py': PythonDocStringDocument,
}


class Sybil:
    """
    An object to provide test runner integration for discovering examples
    in documentation and ensuring they are correct.
    
    :param parsers: 
      A sequence of callable :term:`parsers <parser>`.
      
    :param path:
      The path in which source files are found, relative
      to the path of the Python source file in which this class is instantiated.
      Absolute paths can also be passed.

      .. note::

        This is ignored when using the :ref:`pytest integration <pytest_integration>`.
      
    :param pattern:
      An optional :func:`pattern <fnmatch.fnmatch>` used to match source
      files that will be parsed for examples.
      
    :param patterns:
      An optional sequence of :func:`patterns <fnmatch.fnmatch>` used to match source
      paths that will be parsed for examples.

    :param exclude:
      An optional :func:`pattern <fnmatch.fnmatch>` for source file names
      that will be excluded when looking for examples.

    :param excludes:
      An optional  sequence of :func:`patterns <fnmatch.fnmatch>` for source paths
      that will be excluded when looking for examples.

    :param filenames:
      An optional collection of file names that, if found anywhere within the
      root ``path`` or its sub-directories, will be parsed for examples.

    :param setup:
      An optional callable that will be called once before any examples from
      a :class:`~sybil.document.Document` are evaluated. If provided, it is
      called with the document's :attr:`~sybil.Document.namespace`.
      
    :param teardown:
      An optional callable that will be called after all the examples from
      a :class:`~sybil.document.Document` have been evaluated. If provided, 
      it is called with the document's :attr:`~sybil.Document.namespace`.
      
    :param fixtures:
      An optional sequence of strings specifying the names of fixtures to 
      be requested when using the  :ref:`pytest integration <pytest_integration>`.
      The fixtures will be inserted into the document's :attr:`~sybil.Document.namespace`
      before any examples for that document are evaluated.
      All scopes of fixture are supported.

    :param encoding:
      An optional string specifying the encoding to be used when decoding documentation
      source files.

    :param document_types:
      A mapping of file extension to :class:`Document` subclass such that custom evaluation
      can be performed per document type.

    :param name:
      A name to use in test identifiers so that the identifier indicates which :class:`Sybil`
      that test was discovered by.
    """
    def __init__(
        self,
        parsers: Sequence[Parser],
        pattern: Optional[str] = None,
        patterns: Sequence[str] = (),
        exclude: Optional[str] = None,
        excludes: Sequence[str] = (),
        filenames: Collection[str] = (),
        path: str = '.',
        setup: Optional[Callable[[Dict[str, Any]], None]] = None,
        teardown: Optional[Callable[[Dict[str, Any]], None]] = None,
        fixtures: Sequence[str] = (),
        encoding: str = 'utf-8',
        document_types: Optional[Mapping[Optional[str], Type[Document]]] = None,
        name: str = '',
    ) -> None:

        self.parsers: Sequence[Parser] = parsers
        current_frame = inspect.currentframe()
        calling_frame = current_frame.f_back
        assert calling_frame is not None, 'Cannot find previous frame, which is weird...'
        calling_filename = inspect.getframeinfo(calling_frame).filename
        start_path = Path(calling_filename).parent / path
        self.path: Path = start_path.absolute()
        self.patterns = list(patterns)
        if pattern:
            self.patterns.append(pattern)
        self.excludes = list(excludes)
        if exclude:
            self.excludes.append(exclude)
        self.filenames = filenames
        self.setup: Optional[Callable[[Dict[str, Any]], None]] = setup
        self.teardown: Optional[Callable[[Dict[str, Any]], None]] = teardown
        self.fixtures: Tuple[str, ...] = tuple(fixtures)
        self.encoding: str = encoding
        self.document_types = DEFAULT_DOCUMENT_TYPES.copy()
        if document_types:
            self.document_types.update(document_types)
        self.default_document_type: Type[Document] = self.document_types[None]
        self.name = name

    def __repr__(self) -> str:
        return f'<Sybil: {self.name or str(id(self))}>'

    def __add__(self, other: 'Sybil') -> 'SybilCollection':
        """
        :class:`Sybil` instances can be concatenated into a :class:`~sybil.sybil.SybilCollection`.
        """
        assert isinstance(other, Sybil)
        return SybilCollection((self, other))

    def should_parse(self, path: Path) -> bool:
        try:
            path = path.relative_to(self.path)
        except ValueError:
            return False

        include = False
        if any(path.match(p) for p in self.patterns):
            include = True
        if path.name in self.filenames:
            include = True
        if not include:
            return False

        if any(path.match(e) for e in self.excludes):
            return False
        return True

    def parse(self, path: Path) -> Document:
        type_ = self.document_types.get(path.suffix, self.default_document_type)
        return type_.parse(str(path), *self.parsers, encoding=self.encoding)

    def identify(self, example: Example) -> str:
        sybil_name = f'sybil:{self.name},' if self.name else ''
        return f'{sybil_name}line:{example.line},column:{example.column}'

    def pytest(self) -> Callable[[Path, Any], Any]:
        """
        The helper method for when you use :ref:`pytest_integration`.
        """
        from .integration.pytest import pytest_integration
        return pytest_integration(self)

    def unittest(self) -> Callable[[Any, Any, Optional[str]], Any]:
        """
        The helper method for when you use :ref:`unitttest_integration`.
        """
        from .integration.unittest import unittest_integration
        return unittest_integration(self)


class SybilCollection(List[Sybil]):
    """
    When :class:`Sybil` instances are concatenated, the collection returned can
    be used in the same way as a single :class:`Sybil`.

    This allows multiple configurations to be used in a single test run.
    """

    def pytest(self) -> Callable[[Path, Any], Any]:
        """
        The helper method for when you use :ref:`pytest_integration`.
        """
        from .integration.pytest import pytest_integration
        return pytest_integration(*self)

    def unittest(self) -> Callable[[Any, Any, Optional[str]], Any]:
        """
        The helper method for when you use :ref:`unitttest_integration`.
        """
        from .integration.unittest import unittest_integration
        return unittest_integration(*self)




from sybil.parsers.abstract import AbstractClearNamespaceParser
from .lexers import DirectiveInCommentLexer


class ClearNamespaceParser(AbstractClearNamespaceParser):
    """
    A :any:`Parser` for :ref:`clear-namespace <clear-namespace>` instructions.
    """

    def __init__(self) -> None:
        super().__init__([DirectiveInCommentLexer('clear-namespace')])



import re
import string
from collections.abc import Iterable
from typing import List, Tuple
from textwrap import dedent

from sybil import Region, Document
from sybil.evaluators.capture import evaluate_capture

CAPTURE_DIRECTIVE = re.compile(
    r'^(?P<indent>(\t| )*)\.\.\s*-+>\s*(?P<name>\S+).*$'
)


def indent_matches(line: str, indent: str) -> bool:
    # Is the indentation of a line match what we're looking for?

    if not line.strip():
        # the line consists entirely of whitespace (or nothing at all),
        # so is not considered to be of the appropriate indentation
        return False

    if line.startswith(indent):
        if line[len(indent)] not in string.whitespace:
            return True

    # if none of the above found the indentation to be a match, it is
    # not a match
    return False


class DocumentReversedLines(List[str]):

    def __init__(self, document: Document) -> None:
        super().__init__()
        self[:] = document.text.splitlines(keepends=True)
        self.current_line = len(self)
        self.current_line_end_position = len(document.text)

    def iterate_with_line_number(self) -> Iterable[Tuple[int, str]]:
        while self.current_line > 0:
            self.current_line -= 1
            line = self[self.current_line]
            self.current_line_end_position -= len(line)
            yield self.current_line, line


class CaptureParser:
    """
    A :any:`Parser` for :ref:`captures <capture-parser>`.
    """
    def __call__(self, document: Document) -> Iterable[Region]:
        lines = DocumentReversedLines(document)

        for end_index, line in lines.iterate_with_line_number():

            directive = CAPTURE_DIRECTIVE.match(line)
            if directive:

                region_end = lines.current_line_end_position

                indent = directive.group('indent')
                for start_index, line in lines.iterate_with_line_number():
                    if indent_matches(line, indent):
                        # don't include the preceding line in the capture
                        start_index += 1
                        break
                else:
                    # make it blow up
                    start_index = end_index

                if end_index - start_index < 2:
                    raise ValueError((
                        "couldn't find the start of the block to match "
                        "%r on line %i of %s"
                    ) % (directive.group(), end_index+1, document.path))

                # after dedenting, we need to remove excess leading and trailing
                # newlines, before adding back the final newline that's strippped
                # off
                text = dedent(''.join(lines[start_index:end_index])).strip()+'\n'

                name = directive.group('name')
                parsed = name, text

                yield Region(
                    lines.current_line_end_position,
                    region_end,
                    parsed,
                    evaluate_capture
                )
